--[[
Name: TankPoints
Revision: $Revision: 1 $
Developed by: Whitetooth@Cenarius (hotdogee@巴哈姆特baha)
Description: Calculates and shows your TankPoints in the PaperDall Frame
]]

-----------------
-- Addon Setup --
-----------------

-- Version Info
local MAJOR_VERSION = "2.2.1"
local MINOR_VERSION = tonumber((string.gsub("$Revision: 1 $", "^.-(%d+).-$", "%1")))
-- Librarys
local Compost = AceLibrary("Compost-2.0")
local Gratuity = AceLibrary("Gratuity-2.0")
local Deformat = AceLibrary("Deformat-2.0")
local L = AceLibrary("AceLocale-2.2"):new("TankPoints")
-- AceAddon Initialization
TankPoints = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceConsole-2.0", "AceEvent-2.0", "AceDebug-2.0")
TankPoints.title = "TankPoints"
TankPoints.version = MAJOR_VERSION
--TankPoints.date = string.gsub("$Date: 2006-11-05 15:10:31 -0700 (Sun, 03 Sep 2006) $", "^.-(%d%d%d%d%-%d%d%-%d%d).-$", "%1")

-- Set Debugging
TankPoints:SetDebugging(false)

-------------
-- Globals --
-------------
TP_RANGED = 0
TP_MELEE = 1
TP_HOLY = 2
TP_FIRE = 3
TP_NATURE = 4
TP_FROST = 5
TP_SHADOW = 6
TP_ARCANE = 7

TankPoints.SchoolName = {
	[TP_RANGED] = PLAYERSTAT_RANGED_COMBAT,
	[TP_MELEE] = PLAYERSTAT_MELEE_COMBAT,
	[TP_HOLY] = SPELL_SCHOOL1_CAP,
	[TP_FIRE] = SPELL_SCHOOL2_CAP,
	[TP_NATURE] = SPELL_SCHOOL3_CAP,
	[TP_FROST] = SPELL_SCHOOL4_CAP,
	[TP_SHADOW] = SPELL_SCHOOL5_CAP,
	[TP_ARCANE] = SPELL_SCHOOL6_CAP,
}

---------------------
-- Saved Variables --
---------------------
-- Register DB
TankPoints:RegisterDB("TankPointsDB")
-- Default values
TankPoints.DBDefaults = {
	mobLevelDiff = 3,
	mobDamage = nil,
	mobCritChance = 0.05,
	mobCritBonus = 1,
	mobMissChance = 0.05,
	mobSpellCritChance = 0,
	mobSpellCritBonus = 0.5,
	mobSpellMissChance = 0,
}
-- Register Defaults
TankPoints:RegisterDefaults("profile", TankPoints.DBDefaults)

-- Set Default Mob Stats
function TankPoints.SetDefaultMobStats()
	TankPoints.db.profile.mobLevelDiff = 3
	TankPoints.db.profile.mobDamage = nil
	TankPoints.db.profile.mobCritChance = 0.05
	TankPoints.db.profile.mobCritBonus = 1
	TankPoints.db.profile.mobMissChance = 0.05
	TankPoints.db.profile.mobSpellCritChance = 0
	TankPoints.db.profile.mobSpellCritBonus = 0.5
	TankPoints.db.profile.mobSpellMissChance = 0
	PaperDollFrame_UpdateStats()
	-- Update Calculator
	if TankPointsCalculatorFrame:IsVisible() then
		TPCalc:UpdateBase()
		TPCalc:UpdateResults()
	end
	TankPoints:Print(L["Restored Mob Stats Defaults"])
end

---------------------------
-- Slash Command Options --
---------------------------
local consoleOptions = {
	type = "group",
	args = {
		calc = {
			type = "execute",
			name = L["TankPoints Calculator"],
			desc = L["Shows the TankPoints Calculator"],
			func = function()
				if(TankPointsCalculatorFrame:IsVisible()) then
					TankPointsCalculatorFrame:Hide()
				else
					TankPointsCalculatorFrame:Show()
				end
				PaperDollFrame_UpdateStats()
			end,
		},
		mob = {
			type = "group",
			name = L["Mob Stats"],
			desc = L["Change default mob stats"],
			args = {
				level = {
					type = "range",
					name = L["Mob Level"],
					desc = L["Sets the level difference between the mob and you"],
					get = function() return TankPoints.db.profile.mobLevelDiff end,
					set = function(v) 
						TankPoints.db.profile.mobLevelDiff = v 
						PaperDollFrame_UpdateStats()
						-- Update Calculator
						if TankPointsCalculatorFrame:IsVisible() then
							TPCalc:UpdateBase()
							TPCalc:UpdateResults()
						end
					end,
					min = -20,
					max = 20,
					step = 1,
				},
				damage = {
					type = "range",
					name = L["Mob Damage"],
					desc = L["Sets mob's damage before damage reduction"],
					get = function() return TankPoints:GetMobDamage(UnitLevel("player") + TankPoints.db.profile.mobLevelDiff) end,
					set = function(v)
						TankPoints.db.profile.mobDamage = v
						PaperDollFrame_UpdateStats()
						-- Update Calculator
						if TankPointsCalculatorFrame:IsVisible() then
							TPCalc:UpdateBase()
							TPCalc:UpdateResults()
						end
					end,
					min = 0,
					max = 99999,
					step = 1,
				},
				default = {
					type = "execute",
					name = L["Restore Default"],
					desc = L["Restores default mob stats"],
					func = TankPoints.SetDefaultMobStats,
				},
				advanced = {
					type = "group",
					name = L["Mob Stats Advanced Settings"],
					desc = L["Change advanced mob stats"],
					args = {
						crit = {
							type = "range",
							name = L["Mob Melee Crit"],
							desc = L["Sets mob's melee crit chance"],
							get = function() return TankPoints.db.profile.mobCritChance end,
							set = function(v) 
								TankPoints.db.profile.mobCritChance = v 
								PaperDollFrame_UpdateStats()
								-- Update Calculator
								if TankPointsCalculatorFrame:IsVisible() then
									TPCalc:UpdateBase()
									TPCalc:UpdateResults()
								end
							end,
							min = 0,
							max = 1,
							isPercent = true,
						},
						critbonus = {
							type = "range",
							name = L["Mob Melee Crit Bonus"],
							desc = L["Sets mob's melee crit bonus"],
							get = function() return TankPoints.db.profile.mobCritBonus end,
							set = function(v) 
								TankPoints.db.profile.mobCritBonus = v 
								PaperDollFrame_UpdateStats()
								-- Update Calculator
								if TankPointsCalculatorFrame:IsVisible() then
									TPCalc:UpdateBase()
									TPCalc:UpdateResults()
								end
							end,
							min = 0,
							max = 2,
						},
						miss = {
							type = "range",
							name = L["Mob Melee Miss"],
							desc = L["Sets mob's melee miss chance"],
							get = function() return TankPoints.db.profile.mobMissChance end,
							set = function(v) 
								TankPoints.db.profile.mobMissChance = v 
								PaperDollFrame_UpdateStats()
								-- Update Calculator
								if TankPointsCalculatorFrame:IsVisible() then
									TPCalc:UpdateBase()
									TPCalc:UpdateResults()
								end
							end,
							min = 0,
							max = 1,
							isPercent = true,
						},
						spellcrit = {
							type = "range",
							name = L["Mob Spell Crit"],
							desc = L["Sets mob's spell crit chance"],
							get = function() return TankPoints.db.profile.mobSpellCritChance end,
							set = function(v) 
								TankPoints.db.profile.mobSpellCritChance = v 
								PaperDollFrame_UpdateStats()
								-- Update Calculator
								if TankPointsCalculatorFrame:IsVisible() then
									TPCalc:UpdateBase()
									TPCalc:UpdateResults()
								end
							end,
							min = 0,
							max = 1,
							isPercent = true,
						},
						spellcritbonus = {
							type = "range",
							name = L["Mob Spell Crit Bonus"],
							desc = L["Sets mob's spell crit bonus"],
							get = function() return TankPoints.db.profile.mobSpellCritBonus end,
							set = function(v) 
								TankPoints.db.profile.mobSpellCritBonus = v 
								PaperDollFrame_UpdateStats()
								-- Update Calculator
								if TankPointsCalculatorFrame:IsVisible() then
									TPCalc:UpdateBase()
									TPCalc:UpdateResults()
								end
							end,
							min = 0,
							max = 2,
						},
						spellmiss = {
							type = "range",
							name = L["Mob Spell Miss"],
							desc = L["Sets mob's spell miss chance"],
							get = function() return TankPoints.db.profile.mobSpellMissChance end,
							set = function(v) 
								TankPoints.db.profile.mobSpellMissChance = v 
								PaperDollFrame_UpdateStats()
								-- Update Calculator
								if TankPointsCalculatorFrame:IsVisible() then
									TPCalc:UpdateBase()
									TPCalc:UpdateResults()
								end
							end,
							min = 0,
							max = 1,
							isPercent = true,
						},
					},
				},
			},
		},
	},
}

TankPoints:RegisterChatCommand({"/tp", "/tankpoints"}, consoleOptions)

-----------
-- Tools --
-----------
-- copyTable from AceDB
local function copyTable(to, from)
	setmetatable(to, nil)
	for k,v in pairs(from) do
		if type(k) == "table" then
			k = copyTable({}, k)
		end
		if type(v) == "table" then
			v = copyTable({}, v)
		end
		to[k] = v
	end
	setmetatable(to, from)
	return to
end


---------------------
-- Initializations --
---------------------
function TankPoints:OnInitialize()
	-- Set Dropdown Options Global
	PLAYERSTAT_WARRIOR_STATS = L["TankPoints"]
	-- OnUpdate Frame
	self.OnUpdateFrame = CreateFrame("Frame")
	self.OnUpdateFrame:SetScript("OnUpdate", self.OnUpdate)
	-- Player TankPoints table
	self.playerDataTable = {}
end

function TankPoints:OnEnable()
	-- Add "Warrior Stats" to playerstat drop down list
	tinsert(PLAYERSTAT_DROPDOWN_OPTIONS, "PLAYERSTAT_WARRIOR_STATS")
	-- Hook UpdatePaperdollStats(prefix, index)
	hooksecurefunc("UpdatePaperdollStats", self.UpdatePaperdollStatsHook)
	-- Event to monitor shift, alt, ctrl keys
	self:RegisterEvent("MODIFIER_STATE_CHANGED")
	self:RegisterEvent("UNIT_AURA")
end

function TankPoints:OnDisable()
	-- Remove "Warrior Stats" from playerstat drop down list
	for k, v in ipairs(PLAYERSTAT_DROPDOWN_OPTIONS) do
		if v == "PLAYERSTAT_WARRIOR_STATS" then
			tremove(PLAYERSTAT_DROPDOWN_OPTIONS, k)
			-- reset stats if warrior stats are still selected
			if PLAYERSTAT_LEFTDROPDOWN_SELECTION == v then
				PLAYERSTAT_LEFTDROPDOWN_SELECTION = "PLAYERSTAT_BASE_STATS"
			end
			if PLAYERSTAT_RIGHTDROPDOWN_SELECTION == v then
				local temp, classFileName = UnitClass("player")
				classFileName = strupper(classFileName)
				if classFileName == "MAGE" or classFileName == "PRIEST" or classFileName == "WARLOCK" or classFileName == "DRUID" then
					PLAYERSTAT_RIGHTDROPDOWN_SELECTION = "PLAYERSTAT_SPELL_COMBAT"
				elseif classFileName == "HUNTER" then
					PLAYERSTAT_RIGHTDROPDOWN_SELECTION = "PLAYERSTAT_RANGED_COMBAT"
				else
					PLAYERSTAT_RIGHTDROPDOWN_SELECTION = "PLAYERSTAT_MELEE_COMBAT"
				end
			end
			PaperDollFrame_UpdateStats()
			UIDropDownMenu_SetSelectedValue(PlayerStatFrameLeftDropDown, PLAYERSTAT_LEFTDROPDOWN_SELECTION)
			UIDropDownMenu_SetSelectedValue(PlayerStatFrameRightDropDown, PLAYERSTAT_RIGHTDROPDOWN_SELECTION)
		end
	end
end

------------
-- Events --
------------
-- event = MODIFIER_STATE_CHANGED 
-- arg1 = SHIFT or CTRL or ALT 
-- arg2 = 1 for down, 0 for up 
function TankPoints:MODIFIER_STATE_CHANGED()
	-- Stat Frame Tooltips
	if CharacterAttributesFrame:IsVisible() and GameTooltip:IsVisible() and arg1 == "ALT" then
		self:UpdateStatsTooltips()
	end
	--self:Debug("** "..arg1.." = "..arg2.." **") -- debug
end

function TankPoints:UNIT_AURA()
	-- Do nothing if event target is not player
	if (not arg1 == "player") then
		return
	end
	if PLAYERSTAT_RIGHTDROPDOWN_SELECTION == "PLAYERSTAT_WARRIOR_STATS" then
		--self:Debug("** UNIT_AURA - RIGHT **") -- debug
		TankPoints:Schedule("UpdateWarriorStats".."PlayerStatFrameRight", 0.01, TankPoints.UpdateWarriorStats, TankPoints, "PlayerStatFrameRight", PLAYERSTAT_RIGHTDROPDOWN_SELECTION)
	end
	if PLAYERSTAT_LEFTDROPDOWN_SELECTION == "PLAYERSTAT_WARRIOR_STATS" then
		--self:Debug("** UNIT_AURA - LEFT **") -- debug
		TankPoints:Schedule("UpdateWarriorStats".."PlayerStatFrameLeft", 0.01, TankPoints.UpdateWarriorStats, TankPoints, "PlayerStatFrameLeft", PLAYERSTAT_LEFTDROPDOWN_SELECTION)
	end
end

---------------
-- Schedules --
---------------
-- Schedule Table
TankPoints.ScheduledTasks = {}
function TankPoints:Schedule(taskName, timeAfter, functionName, ...)
	if (taskName and timeAfter) then -- functionName not required so we can use IsScheduled as a timer check
		Compost:Reclaim(self.ScheduledTasks[taskName], 2)
		self.ScheduledTasks[taskName] = Compost:AcquireHash(
			"TargetTime", GetTime() + timeAfter,
			"FunctionName", functionName,
			"Arg", Compost:Acquire(...)
		)
	end
end

function TankPoints:ScheduleRepeat(taskName, repeatRate, functionName, ...)
	if (taskName and repeatRate and functionName) then -- functionName required
		Compost:Reclaim(self.ScheduledTasks[taskName], 2)
		self.ScheduledTasks[taskName] = Compost:AcquireHash(
			"Elapsed", 0,
			"RepeatRate", repeatRate,
			"FunctionName", functionName,
			"Arg", Compost:Acquire(...)
		)
	end
end

function TankPoints:UnSchedule(taskName)
	--WT_RaidWarningAPI.Announce({HOTDOG = taskName})
	if not taskName then return end
	Compost:Reclaim(self.ScheduledTasks[taskName], 2)
	self.ScheduledTasks[taskName] = nil
end

function TankPoints:IsScheduled(taskName)
	return (self.ScheduledTasks[taskName] ~= nil)
end

function TankPoints:OnUpdate(elapsed)
	--TankPoints:Debug("update: "..tostring(elapsed))
	local currentTime = GetTime()
	for taskName, task in pairs(TankPoints.ScheduledTasks) do
		if type(task.TargetTime) ~= "nil" and (currentTime >= task.TargetTime) then
			TankPoints:UnSchedule(taskName)
			if (task.FunctionName) then
				task.FunctionName(unpack(task.Arg))
			end
		elseif type(task.Elapsed) ~= "nil" then
			task.Elapsed = task.Elapsed + arg1
			if (task.Elapsed >= task.RepeatRate) then
				task.FunctionName(unpack(task.Arg))
				task.Elapsed = task.Elapsed - task.RepeatRate
			end
		end
	end
end

---------------------------
-- Character Frame Stats --
---------------------------
function TankPoints.UpdatePaperdollStatsHook(prefix, index)
	-- this gets called multiple times per sec durning stance change or equipment change, which is is not needed
	-- we'll schedule it to update 0.01 sec after its called, fast enough for no delay, slow enough to catch all the repeted calls
	if index == "PLAYERSTAT_WARRIOR_STATS" then
		TankPoints:Schedule("UpdateWarriorStats"..prefix, 0.01, TankPoints.UpdateWarriorStats, TankPoints, prefix, index)
		getglobal(prefix..1):SetScript("OnEnter", TankPoints.TankPointsFrame_OnEnter)
		getglobal(prefix..2):SetScript("OnEnter", TankPoints.MeleeReductionFrame_OnEnter)
		getglobal(prefix..3):SetScript("OnEnter", TankPoints.BlockValueFrame_OnEnter)
		getglobal(prefix..4):SetScript("OnEnter", TankPoints.SpellTankPointsFrame_OnEnter)
		getglobal(prefix..4):SetScript("OnMouseUp", TankPoints.SpellTankPointsFrame_OnMouseUp)
		getglobal(prefix..5):SetScript("OnEnter", TankPoints.SpellReductionFrame_OnEnter)
		getglobal(prefix..5):SetScript("OnMouseUp", TankPoints.SpellTankPointsFrame_OnMouseUp)
		getglobal(prefix..6):SetScript("OnEnter", TankPoints.TankPointsCalculatorStat_OnEnter)
		getglobal(prefix..6):SetScript("OnLeave", TankPoints.TankPointsCalculatorStat_OnLeave)
		getglobal(prefix..6):SetScript("OnMouseUp", TankPoints.ToggleTankPointsCalculator)
	else
		getglobal(prefix..3):SetScript("OnEnter", PaperDollStatTooltip)
		getglobal(prefix..5):SetScript("OnEnter", PaperDollStatTooltip)
		getglobal(prefix..6):SetScript("OnEnter", PaperDollStatTooltip)
		getglobal(prefix..6):SetScript("OnLeave", function (self, motion) GameTooltip:Hide() end)
		getglobal(prefix..4):SetScript("OnMouseUp", nil)
		getglobal(prefix..5):SetScript("OnMouseUp", nil)
		getglobal(prefix..6):SetScript("OnMouseUp", nil)
	end
end

function TankPoints:UpdateWarriorStats(prefix, index)
	-- Calculate TankPoints
	TankPoints:GetPlayerDataTable(self.playerDataTable)
	TankPoints:GetTankPoints(self.playerDataTable)
	-- Set stats
	self:SetTankPoints(getglobal(prefix..1))
	self:SetMeleeReduction(getglobal(prefix..2))
	self:SetBlockValue(getglobal(prefix..3))
	self:SetSpellTankPoints(getglobal(prefix..4))
	self:SetSpellReduction(getglobal(prefix..5))
	self:SetTankPointsCalculator(getglobal(prefix..6))
	-- Update tooltip
	TankPoints:UpdateStatsTooltips()

end

function TankPoints:SetTankPoints(statFrame)
	local tankpoints = format("%d", self.playerDataTable.tankPoints[TP_MELEE])
	PaperDollFrame_SetLabelAndText(statFrame, L["TankPoints"], tankpoints)
	statFrame:Show()
end

function TankPoints:SetMeleeReduction(statFrame)
	local meleeReduction = format("%.2f", self.playerDataTable.totalReduction[TP_MELEE] * 100)
	PaperDollFrame_SetLabelAndText(statFrame, self.SchoolName[TP_MELEE]..L[" DR"], meleeReduction, true)
	statFrame:Show()
end

function TankPoints:SetBlockValue(statFrame)
	local blockValue = self.playerDataTable.blockValue
	PaperDollFrame_SetLabelAndText(statFrame, L["Block Value"], blockValue)
	statFrame:Show()
end

function TankPoints:SetSpellTankPoints(statFrame)
	if TankPoints.setSchool then
		TankPoints.currentSchool = TankPoints.setSchool
	else
		-- Find highest SpellTankPoints school
		TankPoints.currentSchool = TP_FIRE
		for s = TP_FIRE + 1, TP_ARCANE, 1 do
			if self.playerDataTable.tankPoints[s] > self.playerDataTable.tankPoints[TankPoints.currentSchool] then
				TankPoints.currentSchool = s
			end
		end
	end
	local spellTankPoints = format("%d", self.playerDataTable.tankPoints[TankPoints.currentSchool])
	PaperDollFrame_SetLabelAndText(statFrame, self.SchoolName[TankPoints.currentSchool]..L[" TP"], spellTankPoints)
	statFrame:Show()
end

function TankPoints:SetSpellReduction(statFrame)
	if TankPoints.setSchool then
		TankPoints.currentSchool = TankPoints.setSchool
	else
		-- Find highest SpellTankPoints school
		TankPoints.currentSchool = TP_FIRE
		for s = TP_FIRE + 1, TP_ARCANE, 1 do
			if self.playerDataTable.tankPoints[s] > self.playerDataTable.tankPoints[TankPoints.currentSchool] then
				TankPoints.currentSchool = s
			end
		end
	end
	local spellReduction = format("%.2f", self.playerDataTable.totalReduction[TankPoints.currentSchool] * 100)
	PaperDollFrame_SetLabelAndText(statFrame, self.SchoolName[TankPoints.currentSchool]..L[" DR"], spellReduction, true)
	statFrame:Show()
end

function TankPoints:SetTankPointsCalculator(statFrame)
	if(TankPointsCalculatorFrame:IsVisible()) then
		getglobal(statFrame:GetName().."Label"):SetText(L["Close Calculator"])
	else
		getglobal(statFrame:GetName().."Label"):SetText(L["Open Calculator"])
	end
	getglobal(statFrame:GetName().."StatText"):SetText("");
	statFrame:Show()
end

---------------------------
-- TankPoints Calculator --
---------------------------
function TankPoints:ToggleTankPointsCalculator(button)
	if(TankPointsCalculatorFrame:IsVisible()) then
		TankPointsCalculatorFrame:Hide()
		getglobal(self:GetName().."Label"):SetText(L["Open Calculator"])
	else
		TankPointsCalculatorFrame:Show()
		getglobal(self:GetName().."Label"):SetText(L["Close Calculator"])
	end
end

function TankPoints:TankPointsCalculatorStat_OnEnter(motion)
	getglobal(this:GetName().."Label"):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
end

function TankPoints:TankPointsCalculatorStat_OnLeave(motion)
	getglobal(this:GetName().."Label"):SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
end
-----------------
-- Stats Click --
-----------------

-- OnClick Cycle
-- Holy(2) -> Fire(3) -> Nature(4) -> Frost(5) -> Shadow(6) -> Arcane(7)
function TankPoints:SpellTankPointsFrame_OnMouseUp(button)
	-- Set School
	if button == "LeftButton" then
		--TankPoints:Debug(tostring(button))
		if not TankPoints.setSchool then
			TankPoints.setSchool = TankPoints.currentSchool + 1
		else
			TankPoints.setSchool = TankPoints.setSchool + 1
		end
		if TankPoints.setSchool > 7 then
			TankPoints.setSchool = 2
		end
		TankPoints.currentSchool = TankPoints.setSchool
		--[[DEBUG
		local test1 = self:GetScript("OnEnter") == TankPoints.SpellTankPointsFrame_OnEnter
		local test2 = type(self:GetScript("OnEnter"))
		local test3 = self == getglobal("PlayerStatFrameRight"..4)
		TankPoints:Debug(tostring(test1)..", "..tostring(test2)..", "..tostring(test3)..", ")
		--]]
		self:GetScript("OnEnter")(self)
		PaperDollFrame_UpdateStats()
	-- Reset school
	elseif button == "RightButton" then
		TankPoints.setSchool = nil
		-- Find highest SpellTankPoints school
		TankPoints.currentSchool = TP_FIRE
		for s = TP_FIRE + 1, TP_ARCANE, 1 do
			if TankPoints.playerDataTable.tankPoints[s] > TankPoints.playerDataTable.tankPoints[TankPoints.currentSchool] then
				TankPoints.currentSchool = s
			end
		end
		self:GetScript("OnEnter")(self)
		PaperDollFrame_UpdateStats()
	end
end

--------------------
-- Stats Tooltips --
--------------------
--[[
-- Reference --
Font Color Codes:
NORMAL_FONT_COLOR_CODE = "|cffffd200" -- Yellow
HIGHLIGHT_FONT_COLOR_CODE = "|cffffffff" -- White
FONT_COLOR_CODE_CLOSE = "|r"

Font Colors:
NORMAL_FONT_COLOR = {r = 1, g = 0.82, b = 0} -- Yellow
HIGHLIGHT_FONT_COLOR = {r = 1, g = 1, b = 1} -- White

-- TankPoints Tooltip --
in what stance
Mob Stats
Mob Level: 60, Mob Damage: 2000
Mob Crit: 5%, Mob Miss: 5%
1 change in each stat = how many TankPoints
-- Reduction Tooltip --
Armor Damage Reduction
MobLevel, PlayerLevel
Combat Table
-- Block Value --
Mob Damage(Raw)
Mob Damage(DR)
Blocked Percentage
Equivalent Block Mitigation(Block% * BlockMod)
-- Spell TP --
Damage Taken Percentage
25%Melee 75%Spell
50%Melee 50%Spell
75%Melee 25%Spell
-- Spell Reduction --
Improved Defense Stance
Reduction for all schools
--]]
-- TankPoints Stat Tooltip
--[[
-- Static
mobLevel, mobDamage
mobCritChance, mobMissChance

-- Dynamic
-- stat name, increase by statValue, increase by 1
-- strength, 1, 1 -- no strength cause it only increases block value by 0.05
agility, 1, 1
stamina, 1.5, 1
armor, 10, 1
resilience, 1, 1
defenseRating, 1, 1
dodgeRating, 1, 1
parryRating, 1, 1
blockRating, 1, 1
blockValue, 1/0.65, 1
--
TankPoints.MeleePerStatTable = {
	-- stat name, increase by statValue
	-- {SPELL_STAT1_NAME, 1}, -- strength
	{SPELL_STAT2_NAME, 1}, -- agility
	{SPELL_STAT3_NAME, 1.5}, -- stamina
	{ARMOR, 10}, -- armor
	{COMBAT_RATING_NAME15, 1}, -- resilience
	{COMBAT_RATING_NAME2, 10}, -- defenseRating
	{COMBAT_RATING_NAME3, 1}, -- dodgeRating
	{COMBAT_RATING_NAME4, 1}, -- parryRating
	{COMBAT_RATING_NAME5, 1}, -- blockRating
	{L["Block Value"], 1/0.65}, -- blockValue
}
--]]

function TankPoints:UpdateStatsTooltips()
	if PLAYERSTAT_RIGHTDROPDOWN_SELECTION == "PLAYERSTAT_WARRIOR_STATS" then
		for i = 1, 6 do
			local frame = getglobal("PlayerStatFrameRight"..i)
			if GameTooltip:IsOwned(frame) then
				--self:Debug("UpdateStatsTooltips")
				TankPoints:GetPlayerDataTable(self.playerDataTable)
				TankPoints:GetTankPoints(self.playerDataTable)
				local oldthis = this
				this = frame
				frame:GetScript("OnEnter")(frame)
				this = oldthis
				break
			end
		end
	end
	if PLAYERSTAT_LEFTDROPDOWN_SELECTION == "PLAYERSTAT_WARRIOR_STATS" then
		for i = 1, 6 do
			local frame = getglobal("PlayerStatFrameLeft"..i)
			if GameTooltip:IsOwned(frame) then
				TankPoints:GetPlayerDataTable(self.playerDataTable)
				TankPoints:GetTankPoints(self.playerDataTable)
				local oldthis = this
				this = frame
				frame:GetScript("OnEnter")(frame)
				this = oldthis
				break
			end
		end
	end
end

function TankPoints:TankPointsFrame_OnEnter(motion)
	--TankPoints:Debug(motion)
	local time = GetTime() -- Performance Analysis
	local playerDT = TankPoints.playerDataTable
	local textL, textR
	GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
	-------------
	-- Title Line
	textL = L["TankPoints"].." "..format("%d", playerDT.tankPoints[TP_MELEE])
	GameTooltip:SetText(textL, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	---------
	-- Stance
	local currentStance = GetShapeshiftForm()
	if currentStance ~= 0 then
		local _, stanceName = GetShapeshiftFormInfo(currentStance)
		textL = L["In "]..stanceName
		textR = format("%d%%", playerDT.damageTakenMod[TP_MELEE] * 100)
		GameTooltip:AddDoubleLine(textL, textR, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	------------
	-- Mob Stats
	textL = L["Mob Stats"]
	GameTooltip:AddLine(textL, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	-- Mob Level: 60, Mob Damage: 2000
	textL = L["Mob Level"]..": "..playerDT.mobLevel..", "..L["Mob Damage"]..": "..format("%d", playerDT.mobDamage)
	GameTooltip:AddLine(textL, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	-- Mob Crit: 5%, Mob Miss: 5%
	textL = L["Mob Crit"]..": "..format("%.2f", playerDT.mobCritChance * 100).."%, "..L["Mob Miss"]..": "..format("%.2f", playerDT.mobMissChance * 100).."%"
	GameTooltip:AddLine(textL, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	--[[
	TankPoints Per StatValue
	1 Agility = 
	1.5 Stamina = 
	10 Armor = 
	1 Resilience = 
	1 Defense Rating = 
	1 Dodge Rating = 
	1 Parry Rating = 
	1 Block Rating = 
	1/0.65 Block Value = 
	TankPoints.MeleePerStatTable = {
		-- stat name, increase by statValue
		-- {SPELL_STAT1_NAME, 1}, -- strength
		{SPELL_STAT2_NAME, 1}, -- agility
		{SPELL_STAT3_NAME, 1.5}, -- stamina
		{ARMOR, 10}, -- armor
		{COMBAT_RATING_NAME15, 1}, -- resilience
		{COMBAT_RATING_NAME2, 1}, -- defenseRating
		{COMBAT_RATING_NAME3, 1}, -- dodgeRating
		{COMBAT_RATING_NAME4, 1}, -- parryRating
		{COMBAT_RATING_NAME5, 1}, -- blockRating
		{L["Block Value"], 1/0.65}, -- blockValue
	}
	TankPoints:GetPlayerDataTable([TP_Table], [school])
	TankPoints:GetTankPoints([TP_Table], [school])
	--]]
	------------------------------
	-- TankPoints Per StatValue --
	------------------------------
	-- Press Alt key for Per Stat TankPoints
	local alt = IsAltKeyDown()
	if alt then
		textL = L["Per Stat"]
	else
		textL = L["Per StatValue"]
	end
	textR = L["TankPoints"]
	GameTooltip:AddDoubleLine(textL, textR, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	-- Initialize temp melee data table(can't use playerDT because some values have been changed)
	local defaultDT = Compost:Acquire()
	TankPoints:GetPlayerDataTable(defaultDT, TP_MELEE)
	local newDT = Compost:Acquire()

	local changes = Compost:Acquire()
	-------------
	-- Agility --
	-------------
	-- 1 Agi = 2 Armor
	-- 1 Agi = TankPoints:GetDodgePerAgi() dodge% 
	copyTable(newDT, defaultDT) -- load default data

	textL = "1 "..SPELL_STAT2_NAME.." = "
	changes.agi = 1
	TankPoints:AlterDataTable(newDT, changes)
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	-------------
	-- Stamina --
	-------------
	-- 1 Sta = TankPoints:GetHealthPerSta() HP
	copyTable(newDT, defaultDT) -- load default data
	if alt then
		textL = "1 "..SPELL_STAT3_NAME.." = "
		newDT.playerHealth = newDT.playerHealth + 1 * TankPoints:GetHealthPerSta()
	else
		textL = "1.5 "..SPELL_STAT3_NAME.." = "
		newDT.playerHealth = newDT.playerHealth + 1.5 * TankPoints:GetHealthPerSta()
	end
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	-----------
	-- Armor --
	-----------
	copyTable(newDT, defaultDT) -- load default data
	Compost:Recycle(changes)
	if alt then
		textL = "1 "..ARMOR.." = "
		changes.armor = 1
	else	
		textL = "10 "..ARMOR.." = "
		changes.armor = 10
	end
	TankPoints:AlterDataTable(newDT, changes)
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	----------------
	-- Resilience --
	----------------
	copyTable(newDT, defaultDT) -- load default data
	textL = "1 "..COMBAT_RATING_NAME15.." = "
	newDT.resilience = newDT.resilience + 1
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	--------------------
	-- Defense Rating --
	--------------------
	copyTable(newDT, defaultDT) -- load default data
	if alt then
		textL = "1 "..DEFENSE.." = "
		newDT.defense = newDT.defense + 1
		newDT.dodgeChance = newDT.dodgeChance + 0.0004
		newDT.parryChance = newDT.parryChance + 0.0004
		newDT.blockChance = newDT.blockChance + 0.0004
	else
		textL = "1 "..COMBAT_RATING_NAME2.." = "
		newDT.defense = newDT.defense + ReverseRating(1, RB_DEFENSE_SKILL, newDT.playerLevel)
		newDT.dodgeChance = newDT.dodgeChance + 0.0004 * ReverseRating(1, RB_DEFENSE_SKILL, newDT.playerLevel)
		newDT.parryChance = newDT.parryChance + 0.0004 * ReverseRating(1, RB_DEFENSE_SKILL, newDT.playerLevel)
		newDT.blockChance = newDT.blockChance + 0.0004 * ReverseRating(1, RB_DEFENSE_SKILL, newDT.playerLevel)
	end
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	------------------
	-- Dodge Rating --
	------------------
	copyTable(newDT, defaultDT) -- load default data
	if alt then
		textL = "1% "..DODGE.." = "
		newDT.dodgeChance = newDT.dodgeChance + 0.01
	else
		textL = "1 "..COMBAT_RATING_NAME3.." = "
		newDT.dodgeChance = newDT.dodgeChance + ReverseRating(1, RB_DODGE, newDT.playerLevel) * 0.01
	end
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	------------------
	-- Parry Rating --
	------------------
	copyTable(newDT, defaultDT) -- load default data
	if alt then
		textL = "1% "..PARRY.." = "
		newDT.parryChance = newDT.parryChance + 0.01
	else
		textL = "1 "..COMBAT_RATING_NAME4.." = "
		newDT.parryChance = newDT.parryChance + ReverseRating(1, RB_PARRY, newDT.playerLevel) * 0.01
	end
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	------------------
	-- Block Rating --
	------------------
	copyTable(newDT, defaultDT) -- load default data
	if alt then
		textL = "1% "..BLOCK.." = "
		newDT.blockChance = newDT.blockChance + 0.01
	else
		textL = "1 "..COMBAT_RATING_NAME5.." = "
		newDT.blockChance = newDT.blockChance + ReverseRating(1, RB_BLOCK, newDT.playerLevel) * 0.01
	end
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	-----------------
	-- Block Value --
	-----------------
	copyTable(newDT, defaultDT) -- load default data
	if alt then
		textL = "1 "..L["Block Value"].." = "
		newDT.blockValue = newDT.blockValue + 1
	else
		textL = format("%.2f", 1/0.65).." "..L["Block Value"].." = "
		newDT.blockValue = newDT.blockValue + 1/0.65
	end
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	textR = format("%.1f"..L[" TP"], newDT.tankPoints[TP_MELEE] - playerDT.tankPoints[TP_MELEE])
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	------------------------------------
	-- Press ALT for Per Stat TankPoints
	textL = L["Hold ALT for Per Stat TankPoints"]
	GameTooltip:AddLine(textL, GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	
	GameTooltip:Show()
	
	Compost:Reclaim(defaultDT, 2)
	Compost:Reclaim(newDT, 2)
	Compost:Reclaim(changes)
	-- Performance Analysis
	-- Before copyTable: 0.25秒
	-- After copyTable: 0.03秒
	--TankPoints:Debug(format("%.4f", GetTime() - time))
end
--[[
Melee Damage Reduction
Armor Damage Reduction
MobLevel, PlayerLevel
Combat Table
--]]
function TankPoints:MeleeReductionFrame_OnEnter(motion)
	local playerDT = TankPoints.playerDataTable
	local textL, textR
	GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
	-------------
	-- Title Line
	textL = TankPoints.SchoolName[TP_MELEE]..L[" Damage Reduction"].." "..format("%.2f%%", playerDT.totalReduction[TP_MELEE] * 100)
	GameTooltip:SetText(textL, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	
	-------------------------
	-- Armor Damage Reduction
	textL = ARMOR..L[" Damage Reduction"]..":"
	textR = format("%.2f%%", playerDT.armorReduction * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	-- Mob Level: 60, Player Level: 60
	textL = L["Mob Level"]..": "..playerDT.mobLevel..", "..L["Player Level"]..": "..format("%d", playerDT.playerLevel)
	GameTooltip:AddLine(textL, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	---------------
	-- Combat Table
	textL = L["Combat Table"]
	GameTooltip:AddLine(textL, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	
	-- Miss --
	textL = MISS..":"
	textR = format("%.2f%%", playerDT.mobMissChance * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	-- Dodge --
	textL = DODGE..":"
	textR = format("%.2f%%", playerDT.dodgeChance * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	-- Parry --
	textL = PARRY..":"
	textR = format("%.2f%%", playerDT.parryChance * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	-- Block --
	textL = BLOCK..":"
	textR = format("%.2f%%", playerDT.blockChance * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	-- Crit --
	textL = L["Crit"]..":"
	textR = format("%.2f%%", playerDT.mobCritChance * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	-- Crushing --
	textL = L["Crushing"]..":"
	textR = format("%.2f%%", playerDT.mobCrushChance * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	-- Hit --
	textL = L["Hit"]..":"
	textR = format("%.2f%%", (1 - playerDT.mobCrushChance - playerDT.mobCritChance - playerDT.blockChance - playerDT.parryChance - playerDT.dodgeChance - playerDT.mobMissChance) * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	GameTooltip:Show()
end

--[[
Mob Damage(Raw):
Mob Damage(DR):
Blocked Percentage:
Equivalent Block Mitigation:
--]]
function TankPoints:BlockValueFrame_OnEnter(motion)
	local playerDT = TankPoints.playerDataTable
	local textL, textR
	GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
	-------------
	-- Title Line
	textL = L["Block Value"].." "..format("%d", playerDT.blockValue)
	GameTooltip:SetText(textL, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	
	------------------
	-- Mob Damage before DR
	textL = L["Mob Damage before DR"]..":"
	textR = format("%d", TankPoints:GetMobDamage(playerDT.mobLevel))
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	-----------------
	-- Mob Damage after DR
	textL = L["Mob Damage after DR"]..":"
	textR = format("%d", playerDT.mobDamage)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	---------------------
	-- Blocked Percentage
	textL = L["Blocked Percentage"]..":"
	textR = format("%.2f%%", playerDT.blockedMod * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	---------------------
	-- Block Chance
	textL = BLOCK_CHANCE..":"
	textR = format("%.2f%%", playerDT.blockChance * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	------------------------------
	-- Equivalent Block Mitigation
	textL = L["Equivalent Block Mitigation"]..":"
	textR = format("%.2f%%", playerDT.blockChance * playerDT.blockedMod * 100)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	GameTooltip:Show()
end

--[[
TankPoints Considering Melee/Spell Damage Ratio
25% Melee Damage + 75% Spell Damage:
50% Melee Damage + 50% Spell Damage:
75% Melee Damage + 25% Spell Damage:
--]]
function TankPoints:SpellTankPointsFrame_OnEnter(motion)
	--TankPoints:Debug("SpellTankPointsFrame_OnEnter")
	local playerDT = TankPoints.playerDataTable
	local textL, textR
	GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
	-------------------
	-- Spell TankPoints
	textL = TankPoints.SchoolName[TankPoints.currentSchool].." "..L["TankPoints"].." "..format("%d", playerDT.tankPoints[TankPoints.currentSchool])
	GameTooltip:SetText(textL, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	---------
	-- Stance
	local currentStance = GetShapeshiftForm()
	if currentStance ~= 0 then
		local _, stanceName = GetShapeshiftFormInfo(currentStance)
		textL = L["In "]..stanceName
		textR = format("%d%%", playerDT.damageTakenMod[TankPoints.currentSchool] * 100)
		GameTooltip:AddDoubleLine(textL, textR, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	--------------------------------------------------
	-- TankPoints Considering Melee/Spell Damage Ratio
	textL = L["Melee/Spell Damage Ratio"]
	textR = L["TankPoints"]
	GameTooltip:AddDoubleLine(textL, textR, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	--------------------------------------
	-- 25% Melee Damage + 75% Spell Damage
	textL = "25% "..TankPoints.SchoolName[TP_MELEE].." + 75% "..TankPoints.SchoolName[TankPoints.currentSchool]..":"
	textR = format("%d", playerDT.tankPoints[TP_MELEE] * 0.25 + playerDT.tankPoints[TankPoints.currentSchool] * 0.75)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	--------------------------------------
	-- 50% Melee Damage + 50% Spell Damage
	textL = "50% "..TankPoints.SchoolName[TP_MELEE].." + 50% "..TankPoints.SchoolName[TankPoints.currentSchool]..":"
	textR = format("%d", playerDT.tankPoints[TP_MELEE] * 0.50 + playerDT.tankPoints[TankPoints.currentSchool] * 0.50)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	--------------------------------------
	-- 75% Melee Damage + 25% Spell Damage
	textL = "75% "..TankPoints.SchoolName[TP_MELEE].." + 25% "..TankPoints.SchoolName[TankPoints.currentSchool]..":"
	textR = format("%d", playerDT.tankPoints[TP_MELEE] * 0.75 + playerDT.tankPoints[TankPoints.currentSchool] * 0.25)
	GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	
	---------------------------------
	-- Left click: Show next school
	textL = L["Left click: Show next school"]
	GameTooltip:AddLine(textL, GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	---------------------------------
	-- Right click: Show strongest school
	textL = L["Right click: Show strongest school"]
	GameTooltip:AddLine(textL, GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	
	-----------------
	-- Current School
	textL = ""
	local color
	if TP_HOLY == TankPoints.currentSchool then
		color = HIGHLIGHT_FONT_COLOR_CODE
	else
		color = GRAY_FONT_COLOR_CODE
	end
	textL = textL..color..TankPoints.SchoolName[TP_HOLY]..GRAY_FONT_COLOR_CODE
	for s = TP_FIRE, TP_ARCANE do
		if s == TankPoints.currentSchool then
			color = HIGHLIGHT_FONT_COLOR_CODE
		else
			color = GRAY_FONT_COLOR_CODE
		end
		textL = textL.."->"..color..TankPoints.SchoolName[s]..GRAY_FONT_COLOR_CODE
	end
	textL = textL..FONT_COLOR_CODE_CLOSE
	GameTooltip:AddLine(textL)
	
	
	GameTooltip:Show()
end

--[[
Improved Defense Stance
Reduction for all schools
--]]
function TankPoints:SpellReductionFrame_OnEnter(motion)
	local playerDT = TankPoints.playerDataTable
	local textL, textR
	GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
	-------------
	-- Title Line
	textL = TankPoints.SchoolName[TankPoints.currentSchool]..L[" Damage Reduction"].." "..format("%.2f%%", playerDT.totalReduction[TankPoints.currentSchool] * 100)
	GameTooltip:SetText(textL, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	----------------------------
	-- Reduction for all schools
	for s = TP_HOLY, TP_ARCANE do
		textL = getglobal("DAMAGE_SCHOOL"..s)
		textR = format("%.2f%%", playerDT.totalReduction[s] * 100)
		GameTooltip:AddDoubleLine(textL, textR, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b);
		GameTooltip:AddTexture("Interface\\PaperDollInfoFrame\\SpellSchoolIcon"..s);
	end
	GameTooltip:Show()
end

---------------------
-- TankPoints Core --
---------------------
--[[
armorReductionTemp = armor / ((85 * levelModifier) + 400)
armorReduction = armorReductionTemp / (armorReductionTemp + 1)
defenseEffect = (defense - attackerLevel * 5) * 0.04 * 0.01
blockValueFromStrength = (strength * 0.05) - 1
blockValue = floor(blockValueFromStrength) + floor((blockValueFromItems + blockValueFromShield) * blockValueMod)
mobDamage = (levelModifier * 55) * meleeTakenMod * (1 - armorReduction)
resilienceEffect = ReverseRating(resilience, playerLevel) * 0.01
mobCritChance = max(0, 0.05 - defenseEffect - resilienceEffect)
mobCritBonus = 1
mobMissChance = max(0, 0.05 + defenseEffect)
mobCrushChance = 0.15 + max(0, (playerLevel * 5 - defense) * 0.02) (if mobLevel is +3)
mobCritDamageMod = max(0, 1 - resilienceEffect * 2)
blockedMod = min(1, blockValue / mobDamage)
mobSpellCritChance = max(0, 0 - resilienceEffect)
mobSpellCritBonus = 0.5
mobSpellMissChance = 0
mobSpellCritDamageMod = max(0, 1 - resilienceEffect * 2)
schoolReduction[SCHOOL] = 0.75 * (resistance[SCHOOL] / (mobLevel * 5))
totalReduction[MELEE] = 1 - ((mobCritChance * (1 + mobCritBonus) * mobCritDamageMod) + (mobCrushChance * 1.5) + (1 - mobCrushChance - mobCritChance - blockChance * blockedMod - parryChance - dodgeChance - mobMissChance)) * (1 - armorReduction) * meleeTakenMod
totalReduction[SCHOOL] = 1 - ((mobSpellCritChance * (1 + mobSpellCritBonus) * mobSpellCritDamageMod) + (1 - mobSpellCritChance - mobSpellMissChance)) * (1 - schoolReduction[SCHOOL]) * spellTakenMod
tankPoints = playerHealth / (1 - totalReduction)
--]]
function TankPoints:GetArmorReduction(armor, attackerLevel)
	local levelModifier = attackerLevel
	if ( levelModifier > 59 ) then
		levelModifier = levelModifier + (4.5 * (levelModifier - 59))
	end
	local temp = armor / (85 * levelModifier + 400)
	local armorReduction = temp / (1 + temp)
	-- caps at 75%
	if armorReduction > 0.75 then
		armorReduction = 0.75
	end
	if armorReduction < 0 then
		armorReduction = 0
	end
	return armorReduction
end

function TankPoints:GetDefenseEffect(defense, attackerLevel)
	return (defense - attackerLevel * 5) * 0.04 * 0.01
end

function TankPoints:GetDefense()
	local base, modifier = UnitDefense("player");
	return base + modifier
end

function TankPoints:PlayerHasShield()
	local _, _, _, _, _, _, itemSubType = GetItemInfo(GetInventoryItemLink("player", 17) or "")
	return itemSubType == L["Shields"]
end

-- Returns your shield block value, Whitetooth@Cenarius (hotdogee@巴哈姆特baha)
function TankPoints:GetBlockValue(strength)
	-- Block from Strength
	-- Talents: Pal, War
	-- (%d+) Block (on shield)
	-- %+(%d+) Block Value (ZG enchant)
	-- Equip: Increases the block value of your shield by (%d+)
	-- Set: Increases the block value of your shield by (%d+)
	-------------------------------------------------------
	-- Get Block Value from shield if shield is equipped --
	-------------------------------------------------------
	Gratuity:SetInventoryItem("player", GetInventorySlotInfo("SecondaryHandSlot"))
	local _, _, blockValueFromShield = Gratuity:Find(L["^(%d+) Block$"], 4, 6, nil, true)
	if not blockValueFromShield then
		return 0 -- doesn't have shield equipped
	end
	-----------------------------------
	-- Get Block Value from strength --
	-----------------------------------
	if not strength then
		_, strength = UnitStat("player", 1)
	end
	local blockValueFromStrength = (strength * 0.05) - 1
	--------------------------------
	-- Get Block Value from items --
	--------------------------------
	local blockValueFromItems = 0
	local scannedSets = Compost:Acquire() -- table to flag scanned sets so we don't scan twice
	for slot = 1, 18 do -- equipped item's slotID 1 to 18
		Gratuity:SetInventoryItem("player", slot) -- set tooltip
		for i = 2, Gratuity:NumLines() do -- loop through lines
			local line = Gratuity:GetLine(i)
			local num
			-- scan for block value
			for _, pattern in ipairs(L["ItemScan"][TP_BLOCKVALUE]) do
				_, _, num = string.find(line, pattern[1])
				if num then
					blockValueFromItems = blockValueFromItems + num
					break
				end
			end
			-- no block value found in this line
			-- scan for set
			if not num then
				local setName, _, setTotal = Deformat(line, ITEM_SET_NAME)
				if setName then
					if not scannedSets[setName] then
						scannedSets[setName] = true -- flag scanned
						for j = i + setTotal + 2, Gratuity:NumLines() do
							-- ITEM_SET_BONUS = 
							-- Check if set bonus is enabled
							line = Deformat(Gratuity:GetLine(j), ITEM_SET_BONUS)
							if line then
								-- scan for block value
								for _, pattern in ipairs(L["ItemScan"][TP_BLOCKVALUE]) do
									_, _, num = string.find(line, pattern[1])
									if num then
										blockValueFromItems = blockValueFromItems + num
										break
									end
								end
							end
						end
					end
					break -- break out to next slot
				end
			end
		end
	end
	Compost:Reclaim(scannedSets)
	-----------------
	-- Talent Mods --
	-----------------
	-- Warrior: Shield Mastery 3/3, +30% Block Value = 3,16
	-- Paladin: Shield Specialization 3/3, +30% Block Value = 2,8
	-- Shaman: Shield Specialization 5/5, +25% Block Value = 2,2
	-- check class
	local blockValueMod = 1
	local _, class = UnitClass("player")
	if class == "WARRIOR" then
		_, _, _, _, rank = GetTalentInfo(3, 16)
		blockValueMod = 1 + rank * 0.1
	elseif class == "PALADIN" then
		_, _, _, _, rank = GetTalentInfo(2, 8)
		blockValueMod = 1 + rank * 0.1
	elseif class == "SHAMAN" then
		_, _, _, _, rank = GetTalentInfo(2, 2)
		blockValueMod = 1 + rank * 0.05
	end
	-----------------------
	-- Final Calculation --
	-----------------------
	-- Tested to be separate floors by Whitetooth@Cenarius (hotdogee@巴哈姆特baha)
	blockValue = floor(blockValueFromStrength) + floor((blockValueFromItems + blockValueFromShield) * blockValueMod)
	return blockValue, floor(blockValueFromStrength), blockValueFromItems, blockValueFromShield
end

------------------------------------
-- mobDamage, for factoring in block
-- I designed this formula with the goal to model the normal damage of a raid boss at your level
-- the level modifier was taken from the armor reduction formula to base the level effects
-- at level 63 mobDamage is 4455, this is what Nefarian does before armor reduction
-- at level 73 mobDamage is 6518, which matches TBC raid bosses
function TankPoints:GetMobDamage(mobLevel)
	if TankPoints.db.profile.mobDamage then
		return TankPoints.db.profile.mobDamage
	end
	local levelMod = mobLevel
	if ( levelMod > 59 ) then
		levelMod = levelMod + (4.5 * (levelMod - 59))
	end
	return levelMod * 55 -- this is the value before mitigation, which we will do in GetTankPoints
end

--------------------
-- AgiToDodgeBase --
--------------------
-- Derived by Whitetooth@Cenarius (hotdogee@巴哈姆特baha)
TankPoints.AgiToDodgeBase = {
	["WARRIOR"] = 0.7580,
	["PALADIN"] = 0.6520,
	["DRUID"] = -3.3720,
	["SHAMAN"] = 1.6750,
	["HUNTER"] = -7.9500,
	["ROGUE"] = -0.5900,
	["WARLOCK"] = 2.0350,
	["PRIEST"] = 3.1830,
	["MAGE"] = 3.4575,
}

function TankPoints:GetDodgePerAgi()
	local _, class = UnitClass("player")
	local _, agility = UnitStat("player", 2)
	local dodgeFromAgi = GetDodgeChance() - ReverseRating(GetCombatRating(CR_DODGE), RB_DODGE, UnitLevel("player")) - TankPoints:GetDefenseEffect(self:GetDefense(), UnitLevel("player")) * 100
	return (dodgeFromAgi - self.AgiToDodgeBase[class]) / agility
end

------------------
-- HealthPerSta --
------------------
function TankPoints:GetHealthPerSta()
	local _, race = UnitRace("player")
	if race == "Tauren" then
		local _, stamina = UnitStat("player", 3)
		return 10.5, (math.mod(stamina, 2) / 2)
	end
	return 10, 0
end

----------------------
-- TalantBuffEffect --
----------------------
TankPoints.TalantBuffEffects = {
	["DRUID"] = {
		-- Druid: Survival of the Fittest 3/3 - 2,17
		--        Increases all attributes by 3% and reduces the chance you'll be critically hit by melee attacks by 1%/2%/3%.
		["mobCritChance"] = {
			[1] = {
				[TP_MELEE] = true,
				["tab"] = 2,
				["num"] = 17,
				["rank"] = {
					-0.01, -0.02, -0.03,
				},
			},
		},
		-- Druid: Natural Perfection 3/3 - 3,18
		--        Your critical strike chance with all spells is increased by 3% and melee and ranged critical strikes against you cause 4%/7%/10% less damage.
		["mobCritDamageMod"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				["tab"] = 3,
				["num"] = 18,
				["rank"] = {
					-0.04, -0.07, -0.1,
				},
			},
		},
	},
	["HUNTER"] = {
		-- Hunter: Survival Instincts (Rank 2) - 1,14
		--         Reduces all damage taken by 2%/4%.
		["damageTakenMod"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 1,
				["num"] = 14,
				["rank"] = {
					-0.02, -0.04,
				},
			},
		},
	},
	["MAGE"] = {
		-- Mage: Arctic Winds (Rank 5) - 3,20
		--       Reduces the chance melee and ranged attacks will hit you by 1%/2%/3%/4%/5%.
		["mobMissChance"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				["tab"] = 3,
				["num"] = 20,
				["rank"] = {
					0.01, 0.02, 0.03, 0.04, 0.05,
				},
			},
		},
		-- Mage: Prismatic Cloak (Rank 2) - 1,16
		--       Reduces all damage taken by 2%/4%.
		-- Mage: Playing with Fire (Rank 3) - 2,13
		--       Increases all spell damage caused by 1%/2%/3% and all spell damage taken by 1%/2%/3%.
		-- Mage: Frozen Core (Rank 3) - 3,14
		--       Reduces the damage taken by Frost and Fire effects by 2%/4%/6%.
		["damageTakenMod"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 1,
				["num"] = 16,
				["rank"] = {
					-0.02, -0.04,
				},
			},
			[2] = {
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 2,
				["num"] = 13,
				["rank"] = {
					-0.01, -0.02, -0.03,
				},
			},
			[3] = {
				[TP_FIRE] = true,
				[TP_FROST] = true,
				["tab"] = 3,
				["num"] = 14,
				["rank"] = {
					-0.02, -0.04, -0.06,
				},
			},
		},
	},
	["PALADIN"] = {
		-- Paladin: Divine Purpose 3/3 - 3,20
		--          Melee and ranged critical strikes against you cause 4%/7%/10% less damage.
		["mobCritDamageMod"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				["tab"] = 3,
				["num"] = 20,
				["rank"] = {
					-0.04, -0.07, -0.1,
				},
			},
		},
		-- Paladin: Blessed Life 3/3 - 1,18
		--          All attacks against you have a 4%/7%/10% chance to cause half damage.
		-- Paladin: Ardent Defender 5/5 - 2,19
		--          When you have less than 20% health, all damage taken is reduced by 10%/20%/30%/40%/50%.
		-- Paladin: Improved Resistance Auras 2/2 - 2,13
		--          Your Resistance Auras also reduce spell damage of the appropriate type taken by an additional 3%/5%.
		["damageTakenMod"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 1,
				["num"] = 18,
				["rank"] = {
					-0.02, -0.035, -0.05,
				},
			},
			[2] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 2,
				["num"] = 19,
				["rank"] = {
					-0.1, -0.2, -0.3, -0.4, -0.5,
				},
				["if"] = "UnitHealth('player') / UnitMaxHealth('player') < 20%",
			},
			[3] = {
				[TP_FIRE] = true,
				["tab"] = 2,
				["num"] = 13,
				["rank"] = {
					-0.03, -0.05,
				},
				["buff"] = L["Fire Resistance Aura"],
			},
			[4] = {
				[TP_FROST] = true,
				["tab"] = 2,
				["num"] = 13,
				["rank"] = {
					-0.03, -0.05,
				},
				["buff"] = L["Frost Resistance Aura"],
			},
			[5] = {
				[TP_SHADOW] = true,
				["tab"] = 2,
				["num"] = 13,
				["rank"] = {
					-0.03, -0.05,
				},
				["buff"] = L["Shadow Resistance Aura"],
			},
		},
	},
	["PRIEST"] = {
		-- Priest: Shadow Resilience 2/2 - 3,16
		--         Reduces the chance you'll be critically hit by all spells by 2%/4%.
		["mobSpellCritDamageMod"] = {
			[1] = {
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 3,
				["num"] = 16,
				["rank"] = {
					-0.02, -0.04,
				},
			},
		},
		-- Priest: Spell Warding 5/5 - 2,4
		--         Reduces all spell damage taken by 2%/4%/6%/8%/10%.
		-- Priest: Pain Suppression - Reduces all damage taken by 60% for 8 sec.
		["damageTakenMod"] = {
			[1] = {
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 2,
				["num"] = 4,
				["rank"] = {
					-0.02, -0.04, -0.06, -0.08, -0.1,
				},
			},
			[2] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["rank"] = {
					-0.6,
				},
				["buff"] = L["Pain Suppression"],
			},
		},
	},
	["ROGUE"] = {
		-- Rogue: Sleight of Hand 2/2 - 3,3
		--        Reduces the chance you are critically hit by melee and ranged attacks by 2% and increases the threat reduction of your Feint ability by 20%.
		["mobCritChance"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				["tab"] = 3,
				["num"] = 3,
				["rank"] = {
					-0.02, -0.04,
				},
			},
		},
		-- Rogue: Heightened Senses 2/2 - 3,12
		--        Increases your Stealth detection and reduces the chance you are hit by spells and ranged attacks by 2%/4%.
		["mobSpellMissChance"] = {
			[1] = {
				[TP_RANGED] = true,
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 3,
				["num"] = 12,
				["rank"] = {
					0.02, 0.04,
				},
			},
		},
		-- Rogue: Deadened Nerves 5/5 - 1,19
		--        Decreases all physical damage taken by 1%/2%/3%/4%/5%
		["damageTakenMod"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				["tab"] = 1,
				["num"] = 19,
				["rank"] = {
					-0.01, -0.02, -0.03, -0.04, -0.05,
				},
			},
		},
	},
	["SHAMAN"] = {
		-- Shaman: Elemental Shields 3/3 - 1,18
		--         Reduces the chance you will be critically hit by melee and ranged attacks by 2%/4%/6%.
		-- Shaman: Elemental Warding 3/3 - 1,4
		--         Reduces damage taken from Fire, Frost and Nature effects by 4%/7%/10%.
		["mobCritChance"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				["tab"] = 1,
				["num"] = 18,
				["rank"] = {
					-0.02, -0.04, -0.06,
				},
			},
		},
		["damageTakenMod"] = {
			[1] = {
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				["tab"] = 1,
				["num"] = 14,
				["rank"] = {
					-0.04, -0.07, -0.1,
				},
			},
		},
	},
	["WARLOCK"] = {
		-- Warlock: Demonic Resilience 3/3 - 2,18
		--          Reduces the chance you'll be critically hit by melee and spells by 1%/2%/3% and reduces all damage your summoned demon takes by 15%.
		["mobCritChance"] = {
			[1] = {
				[TP_MELEE] = true,
				["tab"] = 2,
				["num"] = 18,
				["rank"] = {
					-0.01, -0.02, -0.03,
				},
			},
		},
		["mobSpellCritChance"] = {
			[1] = {
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["tab"] = 2,
				["num"] = 18,
				["rank"] = {
					-0.01, -0.02, -0.03,
				},
			},
		},
		-- Warlock: Master Demonologist 5/5 - 2,17
		--          Voidwalker - Reduces physical damage taken by 2%/4%/6%/8%/10%.
		-- Warlock: Soul Link 1/1 - 2,19
		--          When active, 30% of all damage taken by the caster is taken by your Imp, Voidwalker, Succubus, Felhunter or Felguard demon instead. In addition, both the demon and master will inflict 3% more damage. Lasts as long as the demon is active.
		["damageTakenMod"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				["tab"] = 2,
				["num"] = 17,
				["rank"] = {
					-0.02, -0.04, -0.06, -0.08, -0.1,
				},
				["condition"] = "UnitCreatureFamily('pet') == '"..L["Voidwalker"].."'",
			},
			[2] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["rank"] = {
					-0.2,
				},
				["buff"] = L["Soul Link"],
			},
		},
	},
	["WARRIOR"] = {
		-- Warrior: Shield Wall - Reduces the Physical and magical damage taken by the caster by 75% for 10 sec.
		["damageTakenMod"] = {
			[1] = {
				[TP_MELEE] = true,
				[TP_RANGED] = true,
				[TP_HOLY] = true,
				[TP_FIRE] = true,
				[TP_NATURE] = true,
				[TP_FROST] = true,
				[TP_SHADOW] = true,
				[TP_ARCANE] = true,
				["rank"] = {
					-0.75,
				},
				["buff"] = L["Shield Wall"],
			},
		},
	},
}
--[[
{
	["mobCritChance"] = 
	["mobMissChance"] = 
	["mobCritDamageMod"] = 
	["mobSpellCritChance"] = 
	["mobSpellMissChance"] = 
	["mobSpellCritDamageMod"] = 
	["damageTakenMod"] = 
}
--]]
function TankPoints:GetTalantBuffEffect(type, school)
	local _, class = UnitClass("player")
	if not self.TalantBuffEffects[class][type] then
		return 0
	end
	local stat = 0
	for _, talant in ipairs(self.TalantBuffEffects[class][type]) do
		if talant[school] then
				
			if not talant.condition or loadstring("return "..talant.condition)() then
				if not talant.buff or GetPlayerBuffName(talant.buff) then
					if talant.tab and talant.num then
						local _, _, _, _, r = GetTalentInfo(talant.tab, talant.num)
						if r ~= 0 then
							stat = stat + talant.rank[r]
						end
					elseif talant.buff then -- buffs only
						stat = stat + talant.rank[1]
					end
				end
			end
		end
	end
	return stat
end

----------------
-- TankPoints --
----------------
--[[
TankPoints:GetPlayerDataTable([TP_Table], [school])
TankPoints:AlterDataTable(TP_Table, changes)
TankPoints:CheckDataTable(TP_Table, [school])
TankPoints:GetTankPoints([TP_Table], [school])
-- school
TP_RANGED = 0
TP_MELEE = 1
TP_HOLY = 2
TP_FIRE = 3
TP_NATURE = 4
TP_FROST = 5
TP_SHADOW = 6
TP_ARCANE = 7

-- TP_Table Inputs
{
	playerLevel = ,
	playerHealth = ,
	playerClass = ,
	mobLevel = ,
	resilience = ,
	-- Melee
	mobCritChance = 0.05, -- talant effects
	mobCritBonus = 1,
	mobMissChance = 0.05,
	armor = ,
	defense = ,
	dodgeChance = ,
	parryChance = ,
	blockChance = ,
	blockValue = ,
	mobDamage = ,
	mobCritDamageMod = , -- from talants
	-- Spell
	mobSpellCritChance = 0, -- talant effects
	mobSpellCritBonus = 0.5,
	mobSpellMissChance = 0, -- this should change with mobLevel, but we don't have enough data yet
	resistance = {
		[TP_HOLY] = 0,
		[TP_FIRE] = ,
		[TP_NATURE] = ,
		[TP_FROST] = ,
		[TP_SHADOW] = ,
		[TP_ARCANE] = ,
	},
	mobSpellCritDamageMod = , -- from talants
	-- All
	damageTakenMod = {
		[TP_MELEE] = ,
		[TP_HOLY] = ,
		[TP_FIRE] = ,
		[TP_NATURE] = ,
		[TP_FROST] = ,
		[TP_SHADOW] = ,
		[TP_ARCANE] = ,
	},
}
-- TP_Table Output adds calculated fields to the table
{
	resilienceEffect = ,
	-- Melee - Added
	armorReduction = ,
	defenseEffect = ,
	mobCrushChance = ,
	mobCritDamageMod = , -- from resilience
	blockedMod = ,
	-- Melee - Changed
	mobMissChance = ,
	dodgeChance = ,
	parryChance = ,
	blockChance = ,
	mobCritChance = ,
	mobDamage = ,
	-- Spell - Added
	mobSpellCritDamageMod = ,
	-- Spell - Changed
	mobSpellCritChance = ,
	-- Results
	schoolReduction = {
		[TP_MELEE] = , -- armorReduction
		[TP_HOLY] = ,
		[TP_FIRE] = ,
		[TP_NATURE] = ,
		[TP_FROST] = ,
		[TP_SHADOW] = ,
		[TP_ARCANE] = ,
	},
	totalReduction = {
		[TP_MELEE] = ,
		[TP_HOLY] = ,
		[TP_FIRE] = ,
		[TP_NATURE] = ,
		[TP_FROST] = ,
		[TP_SHADOW] = ,
		[TP_ARCANE] = ,
	},
	tankPoints = {
		[TP_MELEE] = ,
		[TP_HOLY] = ,
		[TP_FIRE] = ,
		[TP_NATURE] = ,
		[TP_FROST] = ,
		[TP_SHADOW] = ,
		[TP_ARCANE] = ,
	},
}
--]]
function TankPoints:GetPlayerDataTable(TP_Table, school)
	if not TP_Table then
		-- Acquire temp table
		TP_Table = Compost:Acquire()
	end
	-- Unit
	local unit = "player"
	TP_Table.playerLevel = UnitLevel(unit)
	TP_Table.playerHealth = UnitHealthMax(unit)
	_, TP_Table.playerClass = UnitClass(unit)
	TP_Table.mobLevel = UnitLevel(unit) + self.db.profile.mobLevelDiff
	-- Resilience
	TP_Table.resilience = GetCombatRating(CR_CRIT_TAKEN_MELEE)
	TP_Table.damageTakenMod = Compost:Acquire()
	----------------
	-- Melee Data --
	----------------
	if (not school) or school == TP_MELEE then
		-- Mob's Default Crit and Miss Chance
		TP_Table.mobCritChance = self.db.profile.mobCritChance + TankPoints:GetTalantBuffEffect("mobCritChance", TP_MELEE)
		TP_Table.mobCritBonus = self.db.profile.mobCritBonus
		TP_Table.mobMissChance = self.db.profile.mobMissChance + TankPoints:GetTalantBuffEffect("mobMissChance", TP_MELEE)
		-- Armor
		_, TP_Table.armor = UnitArmor(unit)
		-- Defense
		TP_Table.defense = self:GetDefense()
		-- Dodge, Parry
		-- 2.0.2.6144 includes defense factors in these functions
		TP_Table.dodgeChance = GetDodgeChance() * 0.01-- + TP_Table.defenseEffect
		TP_Table.parryChance = GetParryChance() * 0.01-- + TP_Table.defenseEffect
		-- Block Chance, Block Value
		-- check if player has shield
		if not self:PlayerHasShield() then
			TP_Table.blockChance = 0
			TP_Table.blockValue = 0
		else
			TP_Table.blockChance = GetBlockChance() * 0.01-- + TP_Table.defenseEffect
			TP_Table.blockValue = self:GetBlockValue()
		end
		-- Melee Taken Mod
		-- Defensive Stance = "Interface\\Icons\\Ability_Warrior_DefensiveStance"
		-- Berserker Stance = "Interface\\Icons\\Ability_Racial_Avatar"
		TP_Table.damageTakenMod[TP_MELEE] = 1
		local currentStance = GetShapeshiftForm()
		if currentStance ~= 0 then
			local icon = GetShapeshiftFormInfo(currentStance)
			if icon == "Interface\\Icons\\Ability_Warrior_DefensiveStance" then
				TP_Table.damageTakenMod[TP_MELEE] = 0.9
			elseif icon == "Interface\\Icons\\Ability_Racial_Avatar" then
				TP_Table.damageTakenMod[TP_MELEE] = 1.1
			end
		end
		TP_Table.damageTakenMod[TP_MELEE] = TP_Table.damageTakenMod[TP_MELEE] + TankPoints:GetTalantBuffEffect("damageTakenMod", TP_MELEE)
		-- mobDamage, for factoring in block
		TP_Table.mobDamage = self:GetMobDamage(TP_Table.mobLevel)
		-- mobCritDamageMod from talants
		TP_Table.mobCritDamageMod = 1 + TankPoints:GetTalantBuffEffect("mobCritDamageMod", TP_MELEE)
	end
	----------------
	-- Spell Data --
	----------------
	if (not school) or school > TP_MELEE then
		TP_Table.mobSpellCritChance = self.db.profile.mobSpellCritChance + TankPoints:GetTalantBuffEffect("mobSpellCritChance", TP_HOLY)
		TP_Table.mobSpellCritBonus = self.db.profile.mobSpellCritBonus
		TP_Table.mobSpellMissChance = self.db.profile.mobSpellMissChance + TankPoints:GetTalantBuffEffect("mobSpellMissChance", TP_HOLY)
		-- Resistances
		TP_Table.resistance = Compost:Acquire()
		if not school then
			for s = 2, 6, 1 do
				_, TP_Table.resistance[s + 1] = UnitResistance(unit, s)
			end
			-- Holy Resistance always 0
			TP_Table.resistance[TP_HOLY] = 0
		else
			_, TP_Table.resistance[school] = UnitResistance(unit, school - 1)
		end
		-- Spell Taken Mod
		-- Defensive Stance = "Interface\\Icons\\Ability_Warrior_DefensiveStance"
		-- Berserker Stance = "Interface\\Icons\\Ability_Racial_Avatar"
		local spellTakenMod = 1
		local currentStance = GetShapeshiftForm()
		if currentStance ~= 0 then
			local icon = GetShapeshiftFormInfo(currentStance)
			if icon == "Interface\\Icons\\Ability_Warrior_DefensiveStance" then
				spellTakenMod = 0.9
				-- Warrior: Improved Defensive Stance 3/3, -6% Spell Damage Taken = 3,18
				_, _, _, _, rank = GetTalentInfo(3, 18)
				spellTakenMod = spellTakenMod - rank * 0.02
			elseif icon == "Interface\\Icons\\Ability_Racial_Avatar" then
				spellTakenMod = 1.1
			end
		end
		for s = TP_HOLY, TP_ARCANE, 1 do
			TP_Table.damageTakenMod[s] = spellTakenMod + TankPoints:GetTalantBuffEffect("damageTakenMod", s)
		end
		-- mobSpellCritDamageMod from talants
		TP_Table.mobSpellCritDamageMod = 1 + TankPoints:GetTalantBuffEffect("mobSpellCritDamageMod", TP_HOLY)
	end
	
	------------------
	-- Return table --
	------------------
	return TP_Table
end

--[[
changes = {
	-- player stats
	str = ,
	agi = ,
	sta = ,
	playerHealth = ,
	armor = ,
	defense = ,
	dodgeChance = ,
	parryChance = ,
	blockChance = ,
	blockValue = ,
	resilience = ,
	-- mob stats
	mobLevel = ,
	mobDamage = ,
}
--]]
function TankPoints:AlterDataTable(TP_Table, changes)
	if changes.str and changes.str ~= 0 then
		local _, strength = UnitStat("player", 1)
		-- subtract block value from current strength, add block value from new strength
		TP_Table.blockValue = TP_Table.blockValue - floor((strength * 0.05) - 1) + floor((max(0, strength + changes.str) * 0.05) - 1)
	end
	if changes.agi and changes.agi ~= 0 then
		local armorMod = 1
		if TP_Table.playerClass == "DRUID" then
			local formID = GetShapeshiftForm()
			if formID ~= 0 then
				local _, name = GetShapeshiftFormInfo(formID)
				--Interface\\Icons\\Ability_Racial_BearForm
				if name == L["Dire Bear Form"] then
					armorMod = 2.8
				elseif name == L["Bear Form"] or name == L["Moonkin Form"] then
					armorMod = 4.6
				end
			end
		end
		TP_Table.armor = TP_Table.armor + changes.agi * 2 * armorMod
		TP_Table.dodgeChance = TP_Table.dodgeChance + changes.agi * TankPoints:GetDodgePerAgi() * 0.01
	end
	if changes.sta and changes.sta ~= 0 then
		hps, rounding = self:GetHealthPerSta()
		TP_Table.playerHealth = TP_Table.playerHealth + math.ceil(changes.sta * hps - rounding)
	end
	if changes.playerHealth and changes.playerHealth ~= 0 then
		TP_Table.playerHealth = TP_Table.playerHealth + changes.playerHealth
	end
	if changes.armor and changes.armor ~= 0 then
		local armorMod = 1
		if TP_Table.playerClass == "DRUID" then
			local formID = GetShapeshiftForm()
			if formID ~= 0 then
				local _, name = GetShapeshiftFormInfo(formID)
				--Interface\\Icons\\Ability_Racial_BearForm
				if name == L["Dire Bear Form"] then
					armorMod = 2.8
				elseif name == L["Bear Form"] or name == L["Moonkin Form"] then
					armorMod = 4.6
				end
			end
		end
		TP_Table.armor = TP_Table.armor + changes.armor * armorMod
	end
	if changes.defense and changes.defense ~= 0 then
		TP_Table.defense = TP_Table.defense + changes.defense
		TP_Table.dodgeChance = TP_Table.dodgeChance + changes.defense * 0.0004
		if GetParryChance() ~= 0 then
			TP_Table.parryChance = TP_Table.parryChance + changes.defense * 0.0004
		end
		if self:PlayerHasShield() then
			TP_Table.blockChance = TP_Table.blockChance + changes.defense * 0.0004
		end
	end
	if changes.dodgeChance and changes.dodgeChance ~= 0 then
		TP_Table.dodgeChance = TP_Table.dodgeChance + changes.dodgeChance
	end
	if changes.parryChance and changes.parryChance ~= 0 then
		if GetParryChance() ~= 0 then
			TP_Table.parryChance = TP_Table.parryChance + changes.parryChance
		end
	end
	if changes.blockChance and changes.blockChance ~= 0 then
		if self:PlayerHasShield() then
			TP_Table.blockChance = TP_Table.blockChance + changes.blockChance
		end
	end
	if changes.blockValue and changes.blockValue ~= 0 then
		if self:PlayerHasShield() then
			TP_Table.blockValue = TP_Table.blockValue + changes.blockValue
		end
	end
	if changes.resilience and changes.resilience ~= 0 then
		TP_Table.resilience = TP_Table.resilience + changes.resilience
	end
	if changes.mobLevel and changes.mobLevel ~= 0 then
		TP_Table.mobLevel = TP_Table.mobLevel + changes.mobLevel
	end
	if changes.mobDamage and changes.mobDamage ~= 0 then
		TP_Table.mobDamage = TP_Table.mobDamage + changes.mobDamage
	end
end

function TankPoints:CheckDataTable(TP_Table, school)
	-- Check for nil
	-- Fix values that are below minimum
	if TP_Table.playerLevel == nil then self.Print("playerLevel".."is nil") return 
	else TP_Table.playerLevel = max(1, TP_Table.playerLevel) end
	if TP_Table.playerHealth == nil then self.Print("playerHealth".."is nil") return
	else TP_Table.playerHealth = max(0, TP_Table.playerHealth) end
	if TP_Table.mobLevel == nil then self.Print("mobLevel".."is nil") return
	else TP_Table.mobLevel = max(1, TP_Table.mobLevel) end
	if TP_Table.resilience == nil then self.Print("resilience".."is nil") return
	else TP_Table.resilience = max(0, TP_Table.resilience) end
	-- Melee
	if (not school) or school == TP_MELEE then
		if TP_Table.mobCritChance == nil then self.Print("mobCritChance".."is nil") return
		else TP_Table.mobCritChance = max(0, TP_Table.mobCritChance) end
		if TP_Table.mobCritBonus == nil then self.Print("mobCritBonus".."is nil") return
		else TP_Table.mobCritBonus = max(0, TP_Table.mobCritBonus) end
		if TP_Table.mobMissChance == nil then self.Print("mobMissChance".."is nil") return
		else TP_Table.mobMissChance = max(0, TP_Table.mobMissChance) end
		if TP_Table.armor == nil then self.Print("armor".."is nil") return
		else TP_Table.armor = max(0, TP_Table.armor) end
		if TP_Table.defense == nil then self.Print("defense".."is nil") return
		else TP_Table.defense = max(0, TP_Table.defense) end
		if TP_Table.dodgeChance == nil then self.Print("dodgeChance".."is nil") return
		else TP_Table.dodgeChance = max(0, TP_Table.dodgeChance) end
		if GetParryChance() == 0 then
			TP_Table.parryChance = 0
		end
		if TP_Table.parryChance == nil then self.Print("parryChance".."is nil") return
		else TP_Table.parryChance = max(0, TP_Table.parryChance) end
		if not self:PlayerHasShield() then
			TP_Table.blockChance = 0
			TP_Table.blockValue = 0
		end
		if TP_Table.blockChance == nil then self.Print("blockChance".."is nil") return
		else TP_Table.blockChance = max(0, TP_Table.blockChance) end
		if TP_Table.blockValue == nil then self.Print("blockValue".."is nil") return
		else TP_Table.blockValue = max(0, TP_Table.blockValue) end
		if TP_Table.mobDamage == nil then self.Print("mobDamage".."is nil") return
		else TP_Table.mobDamage = max(0, TP_Table.mobDamage) end
		if TP_Table.damageTakenMod[TP_MELEE] == nil then self.Print("damageTakenMod[TP_MELEE]".."is nil") return
		else TP_Table.damageTakenMod[TP_MELEE] = max(0, TP_Table.damageTakenMod[TP_MELEE]) end
	end
	-- Spell
	if (not school) or school > TP_MELEE then
		if TP_Table.mobSpellCritChance == nil then self.Print("mobSpellCritChance".."is nil") return
		else TP_Table.mobSpellCritChance = max(0, TP_Table.mobSpellCritChance) end
		if TP_Table.mobSpellCritBonus == nil then self.Print("mobSpellCritBonus".."is nil") return
		else TP_Table.mobSpellCritBonus = max(0, TP_Table.mobSpellCritBonus) end
		if TP_Table.mobSpellMissChance == nil then self.Print("mobSpellMissChance".."is nil") return
		else TP_Table.mobSpellMissChance = max(0, TP_Table.mobSpellMissChance) end
		-- Negative resistances don't work anymore?
		if not school then
			for s = TP_HOLY, TP_ARCANE, 1 do
				if TP_Table.resistance[s] == nil then self.Print("resistance["..s.."]".."is nil") return
				else TP_Table.resistance[s] = max(0, TP_Table.resistance[s]) end
				if TP_Table.damageTakenMod[s] == nil then self.Print("damageTakenMod["..s.."]".."is nil") return
				else TP_Table.damageTakenMod[s] = max(0, TP_Table.damageTakenMod[s]) end
			end
		else
			if TP_Table.resistance[school] == nil then self.Print("resistance["..school.."]".."is nil") return
			else TP_Table.resistance[school] = max(0, TP_Table.resistance[school]) end
			if TP_Table.damageTakenMod[school] == nil then self.Print("damageTakenMod["..school.."]".."is nil") return
			else TP_Table.damageTakenMod[school] = max(0, TP_Table.damageTakenMod[school]) end
		end
	end
	return true
end

function TankPoints:GetTankPoints(TP_Table, school)
	-----------------
	-- Aquire Data --
	-----------------
	-- Set true if temp table is created
	local tempTableFlag
	if not TP_Table then
		tempTableFlag = true
		-- Fill table with player values
		TP_Table = TankPoints:GetPlayerDataTable(nil, school)
	end
	------------------
	-- Check Inputs --
	------------------
	if not TankPoints:CheckDataTable(TP_Table, school) then return end
	-----------------
	-- Caculations --
	-----------------
	-- Resilience Mod
	TP_Table.resilienceEffect = ReverseRating(TP_Table.resilience, RB_RESILIENCE, TP_Table.playerLevel) * 0.01
	if (not school) or school == TP_MELEE then
		-- Armor Reduction
		TP_Table.armorReduction = self:GetArmorReduction(TP_Table.armor, TP_Table.mobLevel)
		-- Defense Mod (may return negative)
		TP_Table.defenseEffect = self:GetDefenseEffect(TP_Table.defense, TP_Table.mobLevel)
		-- Mob's Crit, Miss
		TP_Table.mobCritChance = max(0, TP_Table.mobCritChance - TP_Table.defenseEffect - TP_Table.resilienceEffect)
		TP_Table.mobMissChance = max(0, TP_Table.mobMissChance + TP_Table.defenseEffect)
		-- Crushing Blow Chance
		TP_Table.mobCrushChance = 0
		if (TP_Table.mobLevel - TP_Table.playerLevel) > 2 then -- if mob is 3 levels or above crushing blow will happen
			-- minimum 15% and additional 2% every defense point under playerLevel*5
			TP_Table.mobCrushChance = 0.15 + max(0, (TP_Table.playerLevel * 5 - TP_Table.defense) * 0.02)
		end
		-- Mob's Crit Damage Mod
		TP_Table.mobCritDamageMod = max(0, 1 - TP_Table.resilienceEffect * 2)
		-- Mob Damage
		-- blocked value is subtracted from the damage after armor and stance mods are factored in
		TP_Table.mobDamage = TP_Table.mobDamage * TP_Table.damageTakenMod[TP_MELEE] * (1 - TP_Table.armorReduction)
		-- Blocked Damage Percentage (blockedMod)
		-- ex: if mob hits you for 1000 and you block 200 of it, then you avoid 200/1000 = 20% of the damage
		-- this value multiplied by block% can now be treated like dodge and parry except that these avoid 100% of the damage
		TP_Table.blockedMod = min(1, TP_Table.blockValue / TP_Table.mobDamage)
	end
	if (not school) or school > TP_MELEE then
		-- Mob's Spell Crit
		TP_Table.mobSpellCritChance = max(0, TP_Table.mobSpellCritChance - TP_Table.resilienceEffect)
		-- Mob's Spell Crit Damage Mod
		TP_Table.mobSpellCritDamageMod = max(0, 1 - TP_Table.resilienceEffect * 2)
	end
	---------------------
	-- High caps check --
	---------------------
	if (not school) or school == TP_MELEE then
		-- Hit < Crushing < Crit < Block < Parry < Dodge < Miss
		local combatTable = Compost:Acquire()
		-- build total sums
		local total = TP_Table.mobMissChance
		tinsert(combatTable, total)
		total = total + TP_Table.dodgeChance
		tinsert(combatTable, total)
		total = total + TP_Table.parryChance
		tinsert(combatTable, total)
		total = total + TP_Table.blockChance
		tinsert(combatTable, total)
		total = total + TP_Table.mobCritChance
		tinsert(combatTable, total)
		total = total + TP_Table.mobCrushChance
		tinsert(combatTable, total)
		-- check caps
		if combatTable[1] > 1 then
			TP_Table.mobMissChance = 1
		end
		if combatTable[2] > 1 then
			TP_Table.dodgeChance = max(0, 1 - combatTable[1])
		end
		if combatTable[3] > 1 then
			TP_Table.parryChance = max(0, 1 - combatTable[2])
		end
		if combatTable[4] > 1 then
			TP_Table.blockChance = max(0, 1 - combatTable[3])
		end
		if combatTable[5] > 1 then
			TP_Table.mobCritChance = max(0, 1 - combatTable[4])
		end
		if combatTable[6] > 1 then
			TP_Table.mobCrushChance = max(0, 1 - combatTable[5])
		end
		Compost:Reclaim(combatTable)
	end
	if (not school) or school > TP_MELEE then
		-- Hit < Crit < Miss
		local combatTable = Compost:Acquire()
		-- build total sums
		local total = TP_Table.mobSpellMissChance
		tinsert(combatTable, total)
		total = total + TP_Table.mobSpellCritChance
		tinsert(combatTable, total)
		-- check caps
		if combatTable[1] > 1 then
			TP_Table.mobSpellMissChance = 1
		end
		if combatTable[2] > 1 then
			TP_Table.mobSpellCritChance = max(0, 1 - combatTable[1])
		end
		Compost:Reclaim(combatTable)
	end
	
	--self:Debug(TP_Table.mobMissChance, TP_Table.dodgeChance, TP_Table.parryChance, TP_Table.blockChance, TP_Table.mobCritChance, TP_Table.mobCrushChance)
	------------------------
	-- Final Calculations --
	------------------------
	--self:Debug("TankPoints Caculated")
	if type(TP_Table.schoolReduction) ~= "table" then
		TP_Table.schoolReduction = Compost:Acquire()
	end
	if type(TP_Table.totalReduction) ~= "table" then
		TP_Table.totalReduction = Compost:Acquire()
	end
	if type(TP_Table.tankPoints) ~= "table" then
		TP_Table.tankPoints = Compost:Acquire()
	end
	if not school then
		-- School Reduction
		TP_Table.schoolReduction[TP_MELEE] = TP_Table.armorReduction
		-- Total Reduction
		TP_Table.totalReduction[TP_MELEE] = 1 - ((TP_Table.mobCritChance * (1 + TP_Table.mobCritBonus) * TP_Table.mobCritDamageMod) + (TP_Table.mobCrushChance * 1.5) + (1 - TP_Table.mobCrushChance - TP_Table.mobCritChance - TP_Table.blockChance * TP_Table.blockedMod - TP_Table.parryChance - TP_Table.dodgeChance - TP_Table.mobMissChance)) * (1 - TP_Table.armorReduction) * TP_Table.damageTakenMod[TP_MELEE]
		-- Total Reduction without block
		--TP_Table.totalReduction[TP_MELEE] = 1 - ((TP_Table.mobCritChance * 2 * TP_Table.mobCritDamageMod) + (TP_Table.mobCrushChance * 1.5) + (1 - TP_Table.mobCrushChance - TP_Table.mobCritChance - TP_Table.parryChance - TP_Table.dodgeChance - TP_Table.mobMissChance)) * (1 - TP_Table.armorReduction) * TP_Table.meleeTakenMod
		TP_Table.tankPoints[TP_MELEE] = TP_Table.playerHealth / (1 - TP_Table.totalReduction[TP_MELEE])
		for s = TP_HOLY, TP_ARCANE, 1 do
			-- Resistance Reduction = 0.75 (resistance / (mobLevel * 5)) 
			TP_Table.schoolReduction[s] = 0.75 * (TP_Table.resistance[s] / (TP_Table.mobLevel * 5))
			-- Total Reduction
			TP_Table.totalReduction[s] = 1 - ((TP_Table.mobSpellCritChance * (1 + TP_Table.mobSpellCritBonus) * TP_Table.mobSpellCritDamageMod) + (1 - TP_Table.mobSpellCritChance - TP_Table.mobSpellMissChance)) * (1 - TP_Table.schoolReduction[s]) * TP_Table.damageTakenMod[s]
			-- TankPoints
			TP_Table.tankPoints[s] = TP_Table.playerHealth / (1 - TP_Table.totalReduction[s])
		end
	else
		if school == TP_MELEE then
			-- School Reduction
			TP_Table.schoolReduction[school] = TP_Table.armorReduction
			-- Total Reduction
			TP_Table.totalReduction[school] = 1 - ((TP_Table.mobCritChance * (1 + TP_Table.mobCritBonus) * TP_Table.mobCritDamageMod) + (TP_Table.mobCrushChance * 1.5) + (1 - TP_Table.mobCrushChance - TP_Table.mobCritChance - TP_Table.blockChance * TP_Table.blockedMod - TP_Table.parryChance - TP_Table.dodgeChance - TP_Table.mobMissChance)) * (1 - TP_Table.armorReduction) * TP_Table.damageTakenMod[school]
			-- Total Reduction without block
			--TP_Table.totalReduction[TP_MELEE] = 1 - ((TP_Table.mobCritChance * 2 * TP_Table.mobCritDamageMod) + (TP_Table.mobCrushChance * 1.5) + (1 - TP_Table.mobCrushChance - TP_Table.mobCritChance - TP_Table.parryChance - TP_Table.dodgeChance - TP_Table.mobMissChance)) * (1 - TP_Table.armorReduction) * TP_Table.meleeTakenMod
			TP_Table.tankPoints[school] = TP_Table.playerHealth / (1 - TP_Table.totalReduction[school])
		else
			-- Resistance Reduction
			TP_Table.schoolReduction[school] = 0.75 * (TP_Table.resistance[school] / (TP_Table.mobLevel * 5))
			-- Total Reduction
			TP_Table.totalReduction[school] = 1 - ((TP_Table.mobSpellCritChance * (1 + TP_Table.mobSpellCritBonus) * TP_Table.mobSpellCritDamageMod) + (1 - TP_Table.mobSpellCritChance - TP_Table.mobSpellMissChance)) * (1 - TP_Table.schoolReduction[school]) * TP_Table.damageTakenMod[school]
			-- TankPoints
			TP_Table.tankPoints[school] = TP_Table.playerHealth / (1 - TP_Table.totalReduction[school])
		end
	end
	-------------
	-- Cleanup --
	-------------
	if tempTableFlag and school then
		local tankPoints, totalReduction, schoolReduction = TP_Table.tankPoints[school], TP_Table.totalReduction[school], TP_Table.schoolReduction[school]
		Compost:Reclaim(TP_Table, 2)
		return tankPoints, totalReduction, schoolReduction
	end
	return TP_Table
end